;============================================================================
; Dateiname : MiniDMX.s
; Funktion  : DMX-Ausgang fr den PC
; Assembler : AVA v0.3b-0815-E
; MCU       : AT90S2313
; Quarz     : 9,216 MHz
; nderungen: 18.02.2001 - V1.0
; Copyright (C) 2001 Mathias Dzionsko (madz@gmx.de)
; http://www.circuitsonline.net/forum/view/51281
;============================================================================
.nolist
;#include "AVR.inc"
.include "tn2313def.inc"
.list
;----------------------------------------------------------------------------
; Registerdefenitionen

#define Temp5           r31             ; ZH  Temp0-5 knnen beliebig
#define Temp4           r30             ; ZL  verwendet werden
#define Temp3           r27             ; XH
#define Temp2           r26             ; XL
#define Temp1           r25             ; WH
#define Temp0           r24             ; WL

#define WH              r25             ; WH,WL knnen als 16-Bit Register
#define WL              r24             ; verwendet werden  (adiw, sbiw)

#define Arg7            r23             ; Argumente fr Funktionen,
#define Arg6            r22             ; werden auch fr die Rckgabe
#define Arg5            r21             ; von Funktionswerten verwendet
#define Arg4            r20
#define Arg3            r19
#define Arg2            r18
#define Arg1            r17
#define Arg0            r16

#define IntrTemp1       r29             ; YH  Nur in Interruptfunktionen 
#define IntrTemp0       r28             ; YL  verwenden
#define IntrTemp        r3
#define IntrSREG        r2              ; Zwischenspeicher fr das Statusreg.

#define ZeroReg         r1              ; Ist *immer* auf Null gesetzt
#define LPMReg          r0              ; Rckgaberegister fr den LPM-Befehl
#define Temp            r0              ; Beliebig verwendbar
;----------------------------------------------------------------------------
; Makros

#define DELAY16_H(time)  high((((XTAL*time)/1000)-5)/4)
#define DELAY16_L(time)  low((((XTAL*time)/1000)-5)/4)

#define UARTDELAY_H(time) high((XTAL*time)/7000)
#define UARTDELAY_L(time) low((XTAL*time)/7000)
;----------------------------------------------------------------------------
; Konstanten

; Taktfrequenz der MCU in kHz (fr DELAY16_H/L)
#define XTAL    9216

; UARTTimeOut 5ms (in s)
#define UARTTimeOut  5000

; DMX
#define DMX_PORT  PORTB
#define DMX_PIN    PB0
#define DMX_CH    96  ; Anzahl der DMX-Kanle

; Port B Pins
; PB0 - DMX-Ausgang
; PB1 - Switch
; PB2 - LED1
; PB3 - LED2
; PB4 - LED3
#define PORTB_INIT      0b00011011
#define DDRB_INIT       0b00011101

; Port D Pins
; PD0 - RXD
; PD1 - TXD
; PD6 - Triggersignal fr Oszilloskop (zu Testzwecken)
#define PORTD_INIT      0b00000010
#define DDRD_INIT       0b01000010

; UART
#define UBRR_INIT       4  ; 115200 Baud
#define UCR_INIT        ((1<<TXEN)|(1<<RXEN))

; Kommunikationsprotokoll
#define P_BLKSTART  0x5A
#define P_BLKEND  0xA5
#define P_DMXOUT96  0xA0
#define P_ACK    0xC1
;============================================================================
; Variablen
;                seg eram.data

.dseg
DMXData:  .byte DMX_CH

.cseg
;============================================================================
; Interruptvektoren (fr AT90S2313!)

;                seg abs=0 flash.intvectors

                rjmp Reset      ; Reset
                rjmp IntDummy   ; External Interrupt0
                rjmp IntDummy   ; External Interrupt1
                rjmp IntDummy   ; Input Capture1 Interrupt 
                rjmp IntDummy   ; Output Compare1A Interrupt
                rjmp IntDummy   ; Overflow1 Interrupt
                rjmp IntDummy   ; Overflow0 Interrupt
                rjmp IntDummy   ; UART Receive Complete Interrupt
                rjmp IntDummy   ; UART Data Register Empty Interrupt
                rjmp IntDummy   ; UART Transmit Complete Interrupt
                rjmp IntDummy   ; Analog Comparator Interrupt
;============================================================================
; Dummyfunktion fr nicht bentigte Interrupts.

;                seg removable flash.code.intdummy

IntDummy:       reti
;----------------------------------------------------------------------------
; Funktion: Delay16
;   Verzgert den Programmablauf (mit 16-Bit Zhler).
;   Fr den Aufruf 5 Takte hinzurechnen:
;     ldi WH, high(XX)  ; 1
;     ldi WL, low(XX)   ; 1
;     rcall Delay16     ; 3
;   Das Macro DELAY16_L/H(Zeit) rechnet die richtige Anzahl an Wartezyklen
;   selber aus. Die Zeit wird in Mikrosekunden angegeben.
;   Max. Verzgerung: ca. 21ms bei 12MHz Takt
; Parameter:
;   WH:WL = (Wartezyklen-3)/4     (Wartezyklen>=7) 
; Vernderte Register:
;   Temp0 (WL), Temp1 (WH)
; Rckgabe:
;   keine
;                seg removable flash.code.delay16

Delay16: sbiw WL,1               ; 2
         brne Delay16            ; 1/2
         ret                     ; 4
;----------------------------------------------------------------------------
; Funktion: DMXWriteBit
;   Setzt anhand des C-Bits die DMX-Leitung und wartet dann noch die
;   restliche Bit-Lnge ab.
; Parameter:
;   C-Bit = 0->DMX-Leitung auf 0, 1->DMX-Leitung auf 1
; Vernderte Register:
;   Temp0
; Rckgabe:
;   keine

;    seg removable flash.code.dmxwritebit

#define DMX_BITDELAY  7    ; Diese Funktion mu insgesamt auf
           ; 33 Takte kommen.

DMXWriteBit:    brcs DMXWriteBit1        
                nop                     ; 1+1
                cbi DMX_PORT, DMX_PIN   ; +2
                rjmp DMXWriteBit2       ; +2=6 Takte
DMXWriteBit1:   sbi DMX_PORT, DMX_PIN   ; 2+2
                rjmp DMXWriteBit2       ; +2=6 Takte
    
DMXWriteBit2:  ldi Temp0, DMX_BITDELAY  ; 1
DMXWriteBit3:  dec Temp0    ; +1   
    brne DMXWriteBit3  ; +1/+2
    rjmp DMXWriteBit4  ; +2
DMXWriteBit4:  ret      ; +4=9+6*3=27 Takte
;----------------------------------------------------------------------------
; Funktion: DMXWriteByte
;   Sendet ein Datenbyte ber die DMX-Leitung.
;   Die bentigten Takte fr ein DMX-Bit sind 37. Das ergibt 4,0147 s pro
;   Bit. Die Abweichung betrgt dann 0,37 Prozent, welche innerhalb der
;   zulssigen Toleranz liegt.
;   Interrupts drfen nicht zugelassen werden.
; Parameter:
;   Arg0 = Datenbyte
; Vernderte Register:
;   Temp0
; Rckgabe:
;   keine

;    seg removable flash.code.dmxwritebyte
    
DMXWriteByte:  clc                    ; 1      - Start Bit
                rcall DMXWriteBit       ; +3+33=37 Takte

                ror Arg0              ; 1      - Bit 0
                rcall DMXWriteBit       ; +3+33=37 Takte
                ror Arg0               ;        - Bit 1
                rcall DMXWriteBit        
                ror Arg0               ;        - Bit 2
                rcall DMXWriteBit                        
                ror Arg0               ;        - Bit 3
                rcall DMXWriteBit        
                ror Arg0               ;        - Bit 4
                rcall DMXWriteBit        
                ror Arg0               ;        - Bit 5
                rcall DMXWriteBit        
                ror Arg0               ;        - Bit 6
                rcall DMXWriteBit        
                ror Arg0               ;        - Bit 7
                rcall DMXWriteBit        

                sec                     ;        - Stop Bit 1
                rcall DMXWriteBit
                sec                     ;        - Stop Bit 2
                rcall DMXWriteBit
                ret
;----------------------------------------------------------------------------
; Funktion: DMXWriteBlock
;   Sendet einen DMX-Datenblock inkl. RESET und Startbyte. Die Anzahl der
;   Kanle ist mir DMX_CH gegeben. Die Daten liegen in DMXData.
; Parameter:
;   keine
; Vernderte Register:
;   Arg0, Temp0-2,4
; Rckgabe:
;   keine

;    seg removable flash.code.dmxwriteblock
    
DMXWriteBlock:  cbi DMX_PORT, DMX_PIN  ; RESET-Signal senden
    ldi WH, DELAY16_H(100)  ; 100s (min. 88s) warten
    ldi WL, DELAY16_L(100)
    rcall Delay16    
    
    sbi DMX_PORT, DMX_PIN  ; MARK-Signal senden
    ldi WH, DELAY16_H(10)  ; 10s (min. 8s) warten
    ldi WL, DELAY16_L(10)
    rcall Delay16    
    
    clr Arg0           ; Startbyte (0) senden
                rcall DMXWriteByte

                ldi ZL, DMXData    ; DMX Daten senden
                ldi Temp2, DMX_CH  ; Anzahl der Kanle
DMXWriteBlock1:  ld Arg0, Z+
                rcall DMXWriteByte
                dec Temp2
                brne DMXWriteBlock1
    ret    
;----------------------------------------------------------------------------
; Funktion: UARTRead
;   Empfngt ein Byte von der seriellen Schnittstelle.
; Parameter:
;   keine
; Vernderte Register:
;   Temp0 (WL), Temp1 (WH)
; Rckgabe:
;   T-Bit = 0->Arg0 enthlt empfangenes Byte, 1->Fehler (Timeout)
;   Arg0 = empfangenes Byte

;    seg removable flash.code.uartrecv
    
UARTRead:       ldi WH, UARTDELAY_H(UARTTimeOut) ; Timeoutzhler setzen
                ldi WL, UARTDELAY_L(UARTTimeOut)

UARTRead1:      sbis USR, RXC           ; Ist was angekommen?
                rjmp UARTRead2
                in Arg0, UDR          ; Byte lesen
                clt                     ; T-Bit lschen
                ret

UARTRead2:      sbiw WL, 1           ; Zeit berschritten?
                brne UARTRead1          
                set                     ; T-Bit setzen
                ret
;----------------------------------------------------------------------------
; Funktion: UARTReadBlock
;   Empfngt DMX_CH Bytes von der seriellen Schnittstelle und speichert
;   sie ab DMXData ab.
; Parameter:
;   keine
; Vernderte Register:
;   Arg0, Temp0 (WL), Temp1 (WH), Temp2, Temp4 (ZL)
; Rckgabe:
;   T-Bit = 0->kein Fehler aufgetreten, 1->Fehler (Timeout)

UARTReadBlock:  ldi ZL, DMXData    ; Zeiger auf DMXData
    ldi Temp2, DMX_CH  ; Anzahl der Kanle

UARTReadBlock1: rcall UARTRead          ; Byte holen
                brtc UARTReadBlock2     ; Zeitberschreitung?
                ret      ; T-Bit ist gesetzt

UARTReadBlock2: st Z+, Arg0             ; Byte abspeichern
                dec Temp2    ; Weitere Bytes?
                brne UARTReadBlock1
                ret      ; T-Bit ist gelscht
;----------------------------------------------------------------------------
; Funktion: UARTWrite
;   Sendet ein Byte ber den UART.
; Parameter:
;   Arg0 = Zu sendendes Byte
; Vernderte Register:
;   keine
; Rckgabe:
;   keine

;    seg removable flash.code.uartwrite
    
UARTWrite:      sbis USR, UDRE          ; Letztes Byte schon gesendet?
                rjmp UARTWrite
                out UDR, Arg0    ; Byte senden
                ret
;============================================================================
; Hauptprogramm

;                seg removable flash.code.reset

Reset:          ldi Arg0, low(RAMEND)   ; Stackpointer setzen
                out SPL, Arg0

                clr ZeroReg             ; Nullregister lschen

                ldi Arg0, PORTB_INIT    ; Port B initialisieren
                out PORTB, Arg0
                ldi Arg0, DDRB_INIT     
                out DDRB, Arg0

                ldi Arg0, PORTD_INIT    ; Port D initialisieren
                out PORTD, Arg0
                ldi Arg0, DDRD_INIT     
                out DDRD, Arg0

                ldi Arg0, UBRR_INIT      ; UART initialisieren
                out UBRR, Arg0
                ldi Arg0, UCR_INIT
                out UCR, Arg0

    ldi ZL, DMXData    ; alle Kanle auf Null
Reset1:    st Z+, ZeroReg
    cpi ZL, DMXData+DMX_CH
    brlo Reset1

MainLoop:       rcall UARTRead          ; auf Daten von dem UART warten
                brts MainLoop

                cpi Arg0, P_BLKSTART    ; Kommt ein Datenblock?
                brne MainLoop

                rcall UARTRead          ; Blocktyp einlesen
                brts MainLoop

                cpi Arg0, P_DMXOUT96   ; DMX Daten senden?
                brne MainLoop

                rcall UARTReadBlock  ; Kanaldaten einlesen
                brts MainLoop

                rcall UARTRead           ; Stimmt das Blockende?
                brts MainLoop
                cpi Arg0, P_BLKEND
                brne MainLoop

                rcall DMXWriteBlock  ; DMX Daten senden
                
    ldi Arg0, P_BLKSTART    ; Besttigung senden
                rcall UARTWrite
                ldi Arg0, P_ACK
                rcall UARTWrite
                ldi Arg0, P_BLKEND
                rcall UARTWrite
                
    in Arg0, PORTB    ; LED blinken lassen
    ldi Arg1, (1<<PB2)
    eor Arg0, Arg1
    out PORTB, Arg0

                rjmp MainLoop
;============================================================================
